@@ -1,26 +1,13 @@
 /*
- * This file is part of the KDE Libraries
- * Copyright (C) 2000 Timo Hummel <timo.hummel@sap.com>
- *                    Tom Braun <braunt@fh-konstanz.de>
- * Copyright 2009 KDE e.V.
- *   By Adriaan de Groot <groot@kde.org>
- * Copyright (C) 2010 George Kiagiadakis <kiagiadakis.george@gmail.com>
- *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Library General Public
- * License as published by the Free Software Foundation; either
- * version 2 of the License, or (at your option) any later version.
- *
- * This library is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Library General Public License for more details.
- *
- * You should have received a copy of the GNU Library General Public License
- * along with this library; see the file COPYING.LIB.  If not, write to
- * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
- * Boston, MA 02110-1301, USA.
- */
+    This file is part of the KDE Libraries
+    SPDX-FileCopyrightText: 2000 Timo Hummel <timo.hummel@sap.com>
+    SPDX-FileCopyrightText: 2000 Tom Braun <braunt@fh-konstanz.de>
+    SPDX-FileCopyrightText: 2010 George Kiagiadakis <kiagiadakis.george@gmail.com>
+    SPDX-FileCopyrightText: 2009 KDE e.V. <kde-ev-board@kde.org> 
+    SPDX-FileContributor: 2009 Adriaan de Groot <groot@kde.org>
+
+    SPDX-License-Identifier: LGPL-2.0-or-later
+*/
 
 #include "kcrash.h"
 
@@ -40,25 +27,29 @@
 #include <qt_windows.h>
 #endif
 #ifdef Q_OS_LINUX
+#include <sys/poll.h>
 #include <sys/prctl.h>
 #endif
 
-#include <kaboutdata.h>
+#include <KAboutData>
 #include <kstartupinfo.h>
 
+#include <array>
+#include <algorithm>
+#include <memory>
+
 #include <QDebug>
 #include <QGuiApplication>
-#include <QFileInfo>
-#include <QDir>
 #include <QStandardPaths>
 #include <QThread>
 #include <QLibraryInfo>
+#include <QFile>
 
 #include <QLoggingCategory>
 Q_DECLARE_LOGGING_CATEGORY(LOG_KCRASH)
 
 // logging category for this framework, default: log stuff >= info
-Q_LOGGING_CATEGORY(LOG_KCRASH, "org.kde.kcrash", QtInfoMsg)
+Q_LOGGING_CATEGORY(LOG_KCRASH, "kf.crash", QtInfoMsg)
 
 #if HAVE_X11
 #include <qx11info_x11.h>
@@ -90,23 +81,101 @@ namespace KCrash
 {
 KCRASH_EXPORT bool loadedByKdeinit = false;
 void setApplicationFilePath(const QString &filePath);
+        // Create socket path to transfer ptrace scope and open connection
 }
+#ifdef Q_OS_LINUX
+static QByteArray s_socketpath;
+#endif
+
+struct Args
+{
+    ~Args()
+    {
+        clear();
+    }
+
+    void clear()
+    {
+        if (!argc) {
+            return;
+        }
+
+        for (int i = 0; i < argc; ++i) {
+            delete[] argv[i];
+        }
+        delete[] argv;
+
+        argv = nullptr;
+        argc = 0;
+    }
+
+    void resize(int size)
+    {
+        clear();
+        argc = size;
+        argv = new char *[argc + 1];
+        for (int i = 0; i < argc + 1; ++i) {
+            argv[i] = nullptr;
+        }
+    }
+
+    explicit operator bool() const
+    {
+        return argc > 0;
+    }
+
+    int argc = 0;
+    // null-terminated array of null-terminated strings
+    char **argv = nullptr;
+};
 
 static KCrash::HandlerType s_emergencySaveFunction = nullptr;
 static KCrash::HandlerType s_crashHandler = nullptr;
-static char *s_appName = nullptr;
-static char *s_autoRestartCommand = nullptr;
-static char *s_appPath = nullptr;
-static int s_autoRestartArgc = 0;
-static char **s_autoRestartCommandLine = nullptr;
-static char *s_drkonqiPath = nullptr;
-static char *s_kdeinit_socket_file = nullptr;
+static std::unique_ptr<char[]> s_appName;
+static std::unique_ptr<char[]> s_appPath;
+static Args s_autoRestartCommandLine;
+static std::unique_ptr<char[]> s_drkonqiPath;
 static KCrash::CrashFlags s_flags = KCrash::CrashFlags();
 static int s_launchDrKonqi = -1; // -1=initial value 0=disabled 1=enabled
+
+static std::unique_ptr<char[]> s_kcrashErrorMessage;
 Q_GLOBAL_STATIC(KCrash::CoreConfig, s_coreConfig)
 
 static void kcrashInitialize()
 {
+
+    // Static because in some cases this is called multiple times
+    // but if an application had any of the bad cases we always want
+    // to skip the check
+    static bool doAutoInitKCrash = true;
+
+    if (!doAutoInitKCrash) {
+        return;
+    }
+
+    QCoreApplication *app = QCoreApplication::instance();
+    if (!app) {
+        doAutoInitKCrash = false;
+        return;
+    }
+
+    if (!QCoreApplication::startingUp()) {
+        // If the app has already started, this means we're not being run as part of
+        // qt_call_pre_routines, which most probably means that we're being run as part
+        // of KCrash being loaded as part of some plugin of the app, so don't
+        // do any magic
+        doAutoInitKCrash = false;
+        return;
+    }
+
+    if (!QCoreApplication::eventDispatcher()) {
+        // We are called with event dispatcher being null when KCrash is being loaded
+        // through plasma-integration instead of being linked to the app (i.e. QtCreator vs Okular)
+        // For apps that don't link directly to KCrash do not do the magic
+        doAutoInitKCrash = false;
+        return;
+    }
+
     KCrash::initialize();
 }
 Q_COREAPP_STARTUP_FUNCTION(kcrashInitialize)
@@ -128,11 +197,13 @@ void KCrash::initialize()
     }
     const QStringList args = QCoreApplication::arguments();
     if (!qEnvironmentVariableIsSet("KDE_DEBUG")
-        && !qEnvironmentVariableIsSet("KCRASH_AUTO_RESTARTED")) {
+        && !qEnvironmentVariableIsSet("KCRASH_AUTO_RESTARTED")
+        && !qEnvironmentVariableIntValue("RUNNING_UNDER_RR")) {
         // enable drkonqi
         KCrash::setDrKonqiEnabled(true);
     } else {
-        qCDebug(LOG_KCRASH) << "KCrash disabled through environment.";
+        // This loads qtlogging.ini very early which prevents unittests from doing QStandardPaths::setTestModeEnabled(true) in initTestCase()
+        //qCDebug(LOG_KCRASH) << "KCrash disabled through environment.";
     }
 
     if (QCoreApplication::instance()) {
@@ -141,6 +212,13 @@ void KCrash::initialize()
         qWarning() << "This process needs a QCoreApplication instance in order to use KCrash";
     }
 
+#ifdef Q_OS_LINUX
+    // Create socket path to transfer ptrace scope and open connection
+    s_socketpath = QFile::encodeName(
+            QStringLiteral("%1/kcrash_%2").arg(QStandardPaths::writableLocation(QStandardPaths::RuntimeLocation))
+                                          .arg(getpid()));
+#endif
+
     s_coreConfig(); // Initialize.
 }
 
@@ -178,7 +256,7 @@ public:
         startTimer(10000); // 10 s
     }
 protected:
-    void timerEvent(QTimerEvent *event) override 
+    void timerEvent(QTimerEvent *event) override
     {
         if (!s_crashHandler) { // not set meanwhile
             KCrash::setCrashHandler(KCrash::defaultCrashHandler);
@@ -210,22 +288,21 @@ void KCrash::setApplicationFilePath(const QString &filePath)
     const QString appName = filePath.mid(pos + 1);
     const QString appPath = filePath.left(pos); // could be empty, in theory
 
-    s_appName = qstrdup(QFile::encodeName(appName).constData());
-    s_appPath = qstrdup(QFile::encodeName(appPath).constData());
+    s_appName.reset(qstrdup(QFile::encodeName(appName).constData()));
+    s_appPath.reset(qstrdup(QFile::encodeName(appPath).constData()));
 
     // Prepare the auto-restart command
-    delete[] s_autoRestartCommand;
-    s_autoRestartCommand = qstrdup(QFile::encodeName(filePath).constData());
-
     QStringList args = QCoreApplication::arguments();
-    args[0] = filePath; // replace argv[0] with full path above
-    delete[] s_autoRestartCommandLine;
-    s_autoRestartArgc = args.count();
-    s_autoRestartCommandLine = new char *[args.count() + 1];
+    if (args.isEmpty()) { // edge case: tst_QX11Info::startupId does QApplication app(argc, nullptr)...
+        args.append(filePath);
+    } else {
+        args[0] = filePath; // replace argv[0] with full path above
+    }
+
+    s_autoRestartCommandLine.resize(args.count());
     for (int i = 0; i < args.count(); ++i) {
-        s_autoRestartCommandLine[i] = qstrdup(QFile::encodeName(args.at(i)).constData());
+        s_autoRestartCommandLine.argv[i] = qstrdup(QFile::encodeName(args.at(i)).constData());
     }
-    s_autoRestartCommandLine[args.count()] = nullptr;
 }
 
 void KCrash::setDrKonqiEnabled(bool enabled)
@@ -238,7 +315,11 @@ void KCrash::setDrKonqiEnabled(bool enabled)
     if (s_launchDrKonqi && !s_drkonqiPath) {
         // search paths
         const QStringList paths = QStringList()
+#if QT_VERSION < QT_VERSION_CHECK(5, 15, 0)
             << QFile::decodeName(qgetenv("LIBEXEC_PATH")).split(QLatin1Char(':'), QString::SkipEmptyParts) // env var is used first
+#else
+            << QFile::decodeName(qgetenv("LIBEXEC_PATH")).split(QLatin1Char(':'), Qt::SkipEmptyParts) // env var is used first
+#endif
             << QCoreApplication::applicationDirPath() // then look where our application binary is located
             << QLibraryInfo::location(QLibraryInfo::LibraryExecutablesPath) // look where libexec path is (can be set in qt.conf)
             << QFile::decodeName(CMAKE_INSTALL_FULL_LIBEXECDIR); // look at our installation location
@@ -247,7 +328,7 @@ void KCrash::setDrKonqiEnabled(bool enabled)
             qCDebug(LOG_KCRASH) << "Could not find drkonqi in search paths:" << paths;
             s_launchDrKonqi = 0;
         } else {
-            s_drkonqiPath = qstrdup(qPrintable(exec));
+            s_drkonqiPath.reset(qstrdup(qPrintable(exec)));
         }
     }
 
@@ -262,55 +343,9 @@ bool KCrash::isDrKonqiEnabled()
     return s_launchDrKonqi == 1;
 }
 
-// The following functions copy&pasted from kinit/wrapper.cpp :
-// (which copied it from kdeinit/kinit.cpp)
-// Can't use QGuiApplication::platformName() here, there is no app instance.
-#if HAVE_X11 || HAVE_XCB
-static const char* displayEnvVarName_c()
-{
-    return "DISPLAY";
-}
-#endif
-
-// adapted from kdeinit/kinit.cpp
-// WARNING, if you change the socket name, adjust kinit.cpp too
-static const QString generate_socket_file_name()
-{
-
-#if HAVE_X11 || HAVE_XCB
-    QByteArray display = qgetenv(displayEnvVarName_c());
-    if (display.isEmpty()) {
-        fprintf(stderr, "Error: could not determine $%s.\n", displayEnvVarName_c());
-        return QString();
-    }
-    int i;
-    if ((i = display.lastIndexOf('.')) > display.lastIndexOf(':') && i >= 0) {
-        display.truncate(i);
-    }
-
-    display.replace(':', '_');
-#ifdef __APPLE__
-    display.replace('/', '_');
-#endif
-#else
-    // not using a DISPLAY variable; use an empty string instead
-    QByteArray display = "";
-#endif
-    const QString socketFileName = QString::fromLatin1("kdeinit5_%1").arg(QLatin1String(display));
-    return socketFileName;
-}
-
 void
 KCrash::setCrashHandler(HandlerType handler)
 {
-    if (!s_kdeinit_socket_file) {
-        // Prepare this now to avoid mallocs in the crash handler.
-        const QString socketFileName = generate_socket_file_name();
-        QByteArray socketName = QFile::encodeName(QStandardPaths::writableLocation(QStandardPaths::RuntimeLocation) +
-                                QLatin1Char('/') + socketFileName);
-        s_kdeinit_socket_file = qstrdup(socketName.constData());
-    }
-
 #if defined(Q_OS_WIN)
     static LPTOP_LEVEL_EXCEPTION_FILTER s_previousExceptionFilter = NULL;
 
@@ -395,9 +430,9 @@ KCrash::defaultCrashHandler(int sig)
         if (s_emergencySaveFunction) {
             s_emergencySaveFunction(sig);
         }
-        if ((s_flags & AutoRestart) && s_autoRestartCommand) {
+        if ((s_flags & AutoRestart) && s_autoRestartCommandLine) {
             QThread::sleep(1);
-            startProcess(s_autoRestartArgc, const_cast<const char **>(s_autoRestartCommandLine), false);
+            startProcess(s_autoRestartCommandLine.argc, const_cast<const char **>(s_autoRestartCommandLine.argv), false);
         }
         crashRecursionCounter++;
     }
@@ -418,16 +453,16 @@ KCrash::defaultCrashHandler(int sig)
         fprintf(stderr, "KCrash: crashing... crashRecursionCounter = %d\n",
                 crashRecursionCounter);
         fprintf(stderr, "KCrash: Application Name = %s path = %s pid = %lld\n",
-                s_appName ? s_appName : "<unknown>",
-                s_appPath ? s_appPath : "<unknown>", QCoreApplication::applicationPid());
+                s_appName ? s_appName.get() : "<unknown>",
+                s_appPath ? s_appPath.get() : "<unknown>", QCoreApplication::applicationPid());
         fprintf(stderr, "KCrash: Arguments: ");
-        for (int i = 0; s_autoRestartCommandLine[i]; ++i) {
-            fprintf(stderr, "%s ", s_autoRestartCommandLine[i]);
+        for (int i = 0; i < s_autoRestartCommandLine.argc; ++i) {
+            fprintf(stderr, "%s ", s_autoRestartCommandLine.argv[i]);
         }
         fprintf(stderr, "\n");
 #else
         fprintf(stderr, "KCrash: Application '%s' crashing...\n",
-                s_appName ? s_appName : "<unknown>");
+                s_appName ? s_appName.get() : "<unknown>");
 #endif
 
         if (s_launchDrKonqi != 1) {
@@ -442,7 +477,7 @@ KCrash::defaultCrashHandler(int sig)
         int i = 0;
 
         // argument 0 has to be drkonqi
-        argv[i++] = s_drkonqiPath;
+        argv[i++] = s_drkonqiPath.get();
 
         const QByteArray platformName = QGuiApplication::platformName().toUtf8();
         if (!platformName.isEmpty()) {
@@ -463,16 +498,16 @@ KCrash::defaultCrashHandler(int sig)
 #endif
 
         argv[i++] = "--appname";
-        argv[i++] = s_appName ? s_appName : "<unknown>";
+        argv[i++] = s_appName ? s_appName.get() : "<unknown>";
 
         if (loadedByKdeinit) {
             argv[i++] = "--kdeinit";
         }
 
         // only add apppath if it's not NULL
-        if (s_appPath && *s_appPath) {
+        if (s_appPath && s_appPath[0]) {
             argv[i++] = "--apppath";
-            argv[i++] = s_appPath;
+            argv[i++] = s_appPath.get();
         }
 
         // signal number -- will never be NULL
@@ -515,7 +550,7 @@ KCrash::defaultCrashHandler(int sig)
             argv[i++] = "--safer";
         }
 
-        if ((s_flags & AutoRestart) && s_autoRestartCommand) {
+        if ((s_flags & AutoRestart) && s_autoRestartCommandLine) {
             argv[i++] = "--restarted";    //tell drkonqi if the app has been restarted
         }
 
@@ -609,131 +644,79 @@ LONG WINAPI KCrash::win32UnhandledExceptionFilter(_EXCEPTION_POINTERS *exception
 }
 #else
 
-static bool startProcessInternal(int argc, const char *argv[], bool waitAndExit, bool directly);
-static pid_t startFromKdeinit(int argc, const char *argv[]);
 static pid_t startDirectly(const char *argv[]);
 static int write_socket(int sock, char *buffer, int len);
 static int read_socket(int sock, char *buffer, int len);
-static int openSocket();
+
+#ifdef Q_OS_LINUX
+static int openDrKonqiSocket(const QByteArray &socketpath);
+static int pollDrKonqiSocket(pid_t pid, int sockfd);
+#endif
 
 void KCrash::startProcess(int argc, const char *argv[], bool waitAndExit)
 {
-    bool startDirectly = true;
+    Q_UNUSED(argc);
+    fprintf(stderr, "KCrash: Attempting to start %s\n", argv[0]);
 
-#ifndef Q_OS_OSX
-    // First try to start the app via kdeinit, if the AlwaysDirectly flag hasn't been specified.
-    // This is done because it is dangerous to use fork() in the crash handler
-    // (there can be functions registered to be performed before fork(), for example handling
-    // of malloc locking, which doesn't work when malloc crashes because of heap corruption).
-    if (!(s_flags & AlwaysDirectly)) {
-        startDirectly = !startProcessInternal(argc, argv, waitAndExit, false);
-    }
-#endif
-
-    // If we can't reach kdeinit, we can still at least try to fork()
-    if (startDirectly) {
-        startProcessInternal(argc, argv, waitAndExit, true);
-    }
-}
-
-static bool startProcessInternal(int argc, const char *argv[], bool waitAndExit, bool directly)
-{
-    fprintf(stderr, "KCrash: Attempting to start %s %s\n", argv[0], directly ? "directly" : "from kdeinit");
-
-    pid_t pid = directly ? startDirectly(argv) : startFromKdeinit(argc, argv);
+    pid_t pid = startDirectly(argv);
 
     if (pid > 0 && waitAndExit) {
         // Seems we made it....
-        alarm(0); //stop the pending alarm that was set at the top of the defaultCrashHandler
+        alarm(0); // Stop the pending alarm that was set at the top of the defaultCrashHandler
 
+        bool running = true;
         // Wait forever until the started process exits. This code path is executed
-        // when launching drkonqi. Note that drkonqi will stop this process in the meantime.
-        if (directly) {
-            //if the process was started directly, use waitpid(), as it's a child...
-            while (waitpid(-1, nullptr, 0) != pid) {}
-        } else {
+        // when launching drkonqi. Note that DrKonqi will SIGSTOP this process in the meantime
+        // and only send SIGCONT when it is about to attach a debugger.
 #ifdef Q_OS_LINUX
-            // Declare the process that will be debugging the crashed KDE app (#245529)
+        // Declare the process that will be debugging the crashed KDE app (#245529).
+        // For now that will be DrKonqi, which may ask to transfer the ptrace scope to
+        // a debugger it is not an ancestor of (because it was started via kdeinit or
+        // KProcess::startDetached()) using a socket.
 #ifndef PR_SET_PTRACER
 # define PR_SET_PTRACER 0x59616d61
 #endif
-            prctl(PR_SET_PTRACER, pid, 0, 0, 0);
+        prctl(PR_SET_PTRACER, pid, 0, 0, 0);
+
+        int sockfd = openDrKonqiSocket(s_socketpath);
+
+        if (sockfd >= 0) {
+            // Wait while DrKonqi is running and the socket connection exists
+            // If the process was started directly, use waitpid(), as it's a child...
+            while ((running = waitpid(pid, nullptr, WNOHANG) != pid) && pollDrKonqiSocket(pid, sockfd) >= 0) {}
+            close(sockfd);
+            unlink(s_socketpath.constData());
+        }
 #endif
-            //...else poll its status using kill()
-            while (kill(pid, 0) >= 0) {
-                sleep(1);
-            }
+        if (running) {
+            // If the process was started directly, use waitpid(), as it's a child...
+            while (waitpid(pid, nullptr, 0) != pid) {}
         }
         if (!s_coreConfig->isProcess()) {
             // Only exit if we don't forward to core dumps
             _exit(253);
         }
     }
-
-    return (pid > 0); //return true on success
-}
-
-static pid_t startFromKdeinit(int argc, const char *argv[])
-{
-    int socket = openSocket();
-    if (socket < -1) {
-        return 0;
-    }
-    kcrash_launcher_header header;
-    header.cmd = LAUNCHER_EXEC_NEW;
-    const int BUFSIZE = 8192; // make sure this is big enough
-    const int CWDSIZE = 2000;
-    char buffer[ BUFSIZE + 10 + 24 /*the env var*/ + CWDSIZE+1 ];
-    int pos = 0;
-    long argcl = argc;
-    memcpy(buffer + pos, &argcl, sizeof(argcl));
-    pos += sizeof(argcl);
-    for (int i = 0;
-            i < argc;
-            ++i) {
-        int len = strlen(argv[ i ]) + 1;   // include terminating \0
-        if (pos + len >= BUFSIZE) {
-            fprintf(stderr, "BUFSIZE in KCrash not big enough!\n");
-            return 0;
-        }
-        memcpy(buffer + pos, argv[ i ], len);
-        pos += len;
-    }
-
-    long env = 1; // 1 env var
-    memcpy(buffer + pos, &env, sizeof(env));
-    pos += sizeof(env);
-    static const char s_envVar[] = "KCRASH_AUTO_RESTARTED=1";
-    int len = strlen(s_envVar) + 1;
-    memcpy(buffer + pos, s_envVar, len);
-    pos += len;
-
-    long avoid_loops = 0;
-    memcpy(buffer + pos, &avoid_loops, sizeof(avoid_loops));
-    pos += sizeof(avoid_loops);
-
-    char cwd[CWDSIZE];
-    if (getcwd(cwd, CWDSIZE-1)) {
-        len = strlen(cwd) + 1; // include terminating \0
-        memcpy(buffer + pos, cwd, len);
-        pos += len;
-    }
-
-    header.arg_length = pos;
-    write_socket(socket, (char *) &header, sizeof(header));
-    write_socket(socket, buffer, pos);
-
-    if (read_socket(socket, (char *) &header, sizeof(header)) < 0
-            || header.cmd != LAUNCHER_OK) {
-        return 0;
-    }
-    long pid;
-    read_socket(socket, (char *) &pid, sizeof(pid));
-    return static_cast<pid_t>(pid);
 }
 
+extern "C" char **environ;
 static pid_t startDirectly(const char *argv[])
 {
+    char** environ_end;
+    for(environ_end = environ; *environ_end; ++environ_end) {}
+
+    std::array<const char*, 1024> environ_data; //hope it's big enough
+    if((unsigned)(environ_end - environ) +2 >= environ_data.size()) {
+        fprintf(stderr, "environ_data in KCrash not big enough!\n");
+        return 0;
+    }
+    auto end = std::copy_if(environ, environ_end, environ_data.begin(),
+                            [](const char* s) {
+                              static const char envvar[] = "KCRASH_AUTO_RESTARTED=";
+                              return strncmp(envvar, s, sizeof(envvar)-1) != 0;
+                            });
+    *end++ = "KCRASH_AUTO_RESTARTED=1";
+    *end++ = nullptr;
     pid_t pid = fork();
     switch (pid) {
     case -1:
@@ -747,8 +730,7 @@ static pid_t startDirectly(const char *argv[])
 #ifndef Q_OS_OSX
         closeAllFDs(); // We are in the child now. Close FDs unconditionally.
 #endif
-        setenv("KCRASH_AUTO_RESTARTED", "1", 1);
-        execvp(argv[0], const_cast< char ** >(argv));
+        execve(argv[0], const_cast< char ** >(argv), const_cast<char**> (environ_data.data()));
         fprintf(stderr, "KCrash failed to exec(), errno = %d\n", errno);
         _exit(253);
     default:
@@ -800,36 +782,95 @@ static int read_socket(int sock, char *buffer, int len)
     return 0;
 }
 
-static int openSocket()
+#ifdef Q_OS_LINUX
+
+static int openDrKonqiSocket(const QByteArray &socketpath)
 {
-    struct sockaddr_un server;
-    if (!s_kdeinit_socket_file ||
-        strlen(s_kdeinit_socket_file) > (sizeof(server.sun_path) - 1))
-    {
-        fprintf(stderr, "kcrash: Unable to communicate with kdeinit5, socket name is %s!",
-                (s_kdeinit_socket_file) ? "too long" : "undefined");
+    int sockfd = socket(PF_UNIX, SOCK_STREAM, 0);
+    if (sockfd < 0) {
+        perror("Warning: socket() for communication with DrKonqi failed");
         return -1;
     }
 
-    /*
-     * create the socket stream
-     */
-    int s = socket(PF_UNIX, SOCK_STREAM, 0);
-    if (s < 0) {
-        perror("Warning: socket() failed: ");
+    struct sockaddr_un drkonqi_server;
+    drkonqi_server.sun_family = AF_UNIX;
+
+    if (socketpath.size() >= static_cast<int>(sizeof(drkonqi_server.sun_path))) {
+        fprintf(stderr, "Warning: socket path is too long\n");
+        close(sockfd);
+        return -1;
+    }
+    strcpy(drkonqi_server.sun_path, socketpath.constData());
+
+    unlink(drkonqi_server.sun_path); // remove potential stale socket
+    if (bind(sockfd, (struct sockaddr *)&drkonqi_server, sizeof(drkonqi_server)) < 0) {
+        perror("Warning: bind() for communication with DrKonqi failed");
+        close(sockfd);
+        unlink(drkonqi_server.sun_path);
         return -1;
     }
 
-    server.sun_family = AF_UNIX;
-    strcpy(server.sun_path, s_kdeinit_socket_file);
-    printf("sock_file=%s\n", s_kdeinit_socket_file);
-    kde_socklen_t socklen = sizeof(server);
-    if (connect(s, (struct sockaddr *)&server, socklen) == -1) {
-        perror("Warning: connect() failed: ");
-        close(s);
-        return -1;
-    }
-    return s;
+    listen(sockfd, 1);
+
+    return sockfd;
 }
 
+static int pollDrKonqiSocket(pid_t pid, int sockfd)
+{
+    struct pollfd fd;
+    fd.fd = sockfd;
+    fd.events = POLLIN;
+    int r;
+    do {
+        r = poll(&fd, 1, 1000); // wait for 1 second for a request by DrKonqi
+    } while (r == -1 && errno == EINTR);
+    // only continue if POLLIN event returned
+    if (r == 0) // timeout
+        return 0;
+    else if (r == -1 || !(fd.revents & POLLIN)) // some error
+        return -1;
+
+    static struct sockaddr_un drkonqi_client;
+    static socklen_t cllength = sizeof(drkonqi_client);
+    int clsockfd;
+    do {
+        clsockfd = accept(sockfd, (struct sockaddr *)&drkonqi_client, &cllength);
+    } while (clsockfd == -1 && errno == EINTR);
+    if (clsockfd < 0)
+        return -1;
+
+    // check whether the message is coming from DrKonqi
+    static struct ucred ucred;
+    static socklen_t credlen = sizeof(struct ucred);
+    if (getsockopt(clsockfd, SOL_SOCKET, SO_PEERCRED, &ucred, &credlen) < 0)
+        return -1;
+
+    if (ucred.pid != pid) {
+        fprintf(stderr, "Warning: peer pid does not match DrKonqi pid\n");
+        return -1;
+    }
+
+    // read PID to change ptrace scope
+    static const int msize = 21; // most digits in a 64bit int (+sign +'\0')
+    char msg[msize];
+    if (read_socket(clsockfd, msg, msize) == 0) {
+        int dpid = atoi(msg);
+        prctl(PR_SET_PTRACER, dpid, 0, 0, 0);
+        // confirm change to DrKonqi
+        if (write_socket(clsockfd, msg, msize) == 0) {
+            fprintf(stderr, "KCrash: ptrace access transferred to %s\n", msg);
+        }
+    }
+    close(clsockfd);
+
+    return 1;
+}
+
+#endif
+
 #endif // Q_OS_UNIX
+
+void KCrash::setErrorMessage(const QString &message)
+{
+    s_kcrashErrorMessage.reset(qstrdup(message.toUtf8().constData()));
+}
